@@ -110,6 +110,14 @@ static const struct intel_device_info intel_haswell_info = {
 	.gen = 075,
 };
 
+static const struct intel_device_info intel_broadwell_info = {
+	.gen = 0100,
+};
+
+static const struct intel_device_info intel_cherryview_info = {
+	.gen = 0101,
+};
+
 static const SymTabRec intel_chipsets[] = {
 	{PCI_CHIP_I810,				"i810"},
 	{PCI_CHIP_I810_DC100,			"i810-dc100"},
@@ -144,6 +152,9 @@ static const SymTabRec intel_chipsets[] = {
 	{PCI_CHIP_G41_G,			"G41"},
 	{PCI_CHIP_B43_G,			"B43"},
 	{PCI_CHIP_B43_G1,			"B43"},
+
+	{0, ""},
+
 	{PCI_CHIP_IRONLAKE_D_G,			"HD Graphics"},
 	{PCI_CHIP_IRONLAKE_M_G,			"HD Graphics"},
 	{PCI_CHIP_SANDYBRIDGE_GT1,		"HD Graphics 2000" },
@@ -204,9 +215,49 @@ static const SymTabRec intel_chipsets[] = {
 	{PCI_CHIP_HASWELL_CRW_E_GT1,		"HD Graphics" }, /* ??? */
 	{PCI_CHIP_HASWELL_CRW_E_GT2,		"HD Graphics" }, /* ??? */
 	{PCI_CHIP_HASWELL_CRW_E_GT3,		"Iris(TM) Pro Graphics 5200" },
-	{-1,					NULL}
+
+	/* Valleyview (Baytail) */
+	{0x0f30, "HD Graphics"},
+	{0x0f31, "HD Graphics"},
+	{0x0f32, "HD Graphics"},
+	{0x0f33, "HD Graphics"},
+	{0x0155, "HD Graphics"},
+	{0x0157, "HD Graphics"},
+
+	/* Broadwell Marketing names */
+	{0x1602, "HD graphics"},
+	{0x1606, "HD graphics"},
+	{0x160B, "HD graphics"},
+	{0x160A, "HD graphics"},
+	{0x160D, "HD graphics"},
+	{0x160E, "HD graphics"},
+	{0x1612, "HD graphics 5600"},
+	{0x1616, "HD graphics 5500"},
+	{0x161B, "HD graphics"},
+	{0x161A, "HD graphics"},
+	{0x161D, "HD graphics"},
+	{0x161E, "HD graphics 5300"},
+	{0x1622, "Iris Pro graphics 6200"},
+	{0x1626, "HD graphics 6000"},
+	{0x162B, "Iris graphics 6100"},
+	{0x162A, "Iris Pro graphics P6300"},
+	{0x162D, "HD graphics"},
+	{0x162E, "HD graphics"},
+	{0x1632, "HD graphics"},
+	{0x1636, "HD graphics"},
+	{0x163B, "HD graphics"},
+	{0x163A, "HD graphics"},
+	{0x163D, "HD graphics"},
+	{0x163E, "HD graphics"},
+
+	/* When adding new identifiers, also update:
+	 * 1. intel_identify()
+	 * 2. man/intel.man
+	 * 3. README
+	 */
+
+	{-1, NULL} /* Sentinel */
 };
-#define NUM_CHIPSETS (sizeof(intel_chipsets) / sizeof(intel_chipsets[0]))
 
 static const struct pci_id_match intel_device_match[] = {
 #if UMS
@@ -218,7 +269,7 @@ static const struct pci_id_match intel_device_match[] = {
 
 #if KMS
 	INTEL_I830_IDS(&intel_i830_info),
-	INTEL_I845G_IDS(&intel_i830_info),
+	INTEL_I845G_IDS(&intel_i845_info),
 	INTEL_I85X_IDS(&intel_i855_info),
 	INTEL_I865G_IDS(&intel_i865_info),
 
@@ -251,6 +302,11 @@ static const struct pci_id_match intel_device_match[] = {
 	INTEL_VLV_D_IDS(&intel_valleyview_info),
 	INTEL_VLV_M_IDS(&intel_valleyview_info),
 
+	INTEL_BDW_D_IDS(&intel_broadwell_info),
+	INTEL_BDW_M_IDS(&intel_broadwell_info),
+
+	INTEL_CHV_IDS(&intel_cherryview_info),
+
 	INTEL_VGA_DEVICE(PCI_MATCH_ANY, &intel_generic_info),
 #endif
 
@@ -334,7 +390,7 @@ static void intel_identify(int flags)
 	xf86Msg(X_INFO, INTEL_NAME ": Driver for Intel(R) Integrated Graphics Chipsets:\n\t");
 	len = 8;
 
-	for (chipset = intel_chipsets; chipset->name; chipset++) {
+	for (chipset = intel_chipsets; chipset->token; chipset++) {
 		for (j = i; --j >= 0;)
 			if (strcmp(unique[j], chipset->name) == 0)
 				break;
@@ -376,6 +432,10 @@ static void intel_identify(int flags)
 	xf86ErrorF("\n");
 	if (unique != stack)
 		free(unique);
+
+	xf86Msg(X_INFO, INTEL_NAME ": Driver for Intel(R) HD Graphics: 2000-6000\n");
+	xf86Msg(X_INFO, INTEL_NAME ": Driver for Intel(R) Iris(TM) Graphics: 5100, 6100\n");
+	xf86Msg(X_INFO, INTEL_NAME ": Driver for Intel(R) Iris(TM) Pro Graphics: 5200, 6200, P6300\n");
 }
 
 static Bool intel_driver_func(ScrnInfoPtr pScrn,
@@ -397,6 +457,12 @@ static Bool intel_driver_func(ScrnInfoPtr pScrn,
 #endif
 
 		return TRUE;
+
+#if XORG_VERSION_CURRENT >= XORG_VERSION_NUMERIC(1,15,99,902,0)
+	case SUPPORTS_SERVER_FDS:
+		return TRUE;
+#endif
+
 	default:
 		/* Unknown or deprecated function */
 		return FALSE;
@@ -419,7 +485,7 @@ _xf86findDriver(const char *ident, XF86ConfDevicePtr p)
 	return NULL;
 }
 
-static enum accel_method { UXA, SNA } get_accel_method(void)
+static enum accel_method { SNA, UXA, GLAMOR } get_accel_method(void)
 {
 	enum accel_method accel_method = DEFAULT_ACCEL_METHOD;
 	XF86ConfDevicePtr dev;
@@ -438,7 +504,7 @@ static enum accel_method { UXA, SNA } get_accel_method(void)
 			else if (strcasecmp(s, "uxa") == 0)
 				accel_method = UXA;
 			else if (strcasecmp(s, "glamor") == 0)
-				accel_method = UXA;
+				accel_method = GLAMOR;
 		}
 	}
 
@@ -477,7 +543,7 @@ intel_scrn_create(DriverPtr		driver,
 	scrn->driverVersion = INTEL_VERSION;
 	scrn->driverName = (char *)INTEL_DRIVER_NAME;
 	scrn->name = (char *)INTEL_NAME;
-	scrn->driverPrivate = (void *)(match_data | 1);
+	scrn->driverPrivate = (void *)(match_data | (flags & XF86_ALLOCATE_GPU_SCREEN) | 2);
 	scrn->Probe = NULL;
 
 	if (xf86IsEntitySharable(entity_num))
@@ -495,7 +561,9 @@ intel_scrn_create(DriverPtr		driver,
 	case SNA: return sna_init_scrn(scrn, entity_num);
 #endif
 #if USE_UXA
-	case UXA: return intel_init_scrn(scrn);
+	case GLAMOR:
+	case UXA:
+		  return intel_init_scrn(scrn);
 #endif
 
 	default: break;
